/*  cat.c
 *
 *  Yaesu FT847/FT857 CAT control functions for the xfhell application
 */

/*
 *  xfhell: An application to transmit and receive
 *  Feld Hell signals using a computer's sound card
 *
 *  Copyright (C) 2003  Neoklis Kyriazis
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of
 *  the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details:
 *
 *  http://www.gnu.org/copyleft/gpl.txt
 */

#include <termios.h>
#include "xfhell.h"
#include "support.h"

/* Average bin values */
extern int gbl_bin_ave[];

/* Runtime config data */
extern rc_data_t rc_data;

/* Command codes for FT847 CAT */
unsigned char
  CAT_ON[]		= { 0x00, 0, 0, 0, 0x00 },
  CAT_OFF[]		= { 0x00, 0, 0, 0, 0x80 },
  PTT_ON[]		= { 0x00, 0, 0, 0, 0x08 },
  PTT_OFF[]		= { 0x00, 0, 0, 0, 0x88 },
  MODE_LSB[]	= { 0x00, 0, 0, 0, 0x07 },
  MODE_USB[]	= { 0x01, 0, 0, 0, 0x07 },
  MODE_CW[]		= { 0x02, 0, 0, 0, 0x07 },
  MODE_CWR[]	= { 0x03, 0, 0, 0, 0x07 },
  MODE_CWN[]	= { 0x82, 0, 0, 0, 0x07 },
  MODE_CWNR[]	= { 0x83, 0, 0, 0, 0x07 },
  MODE_DIG[]	= { 0x0a, 0, 0, 0, 0x07 },
  MODE_PKT[]	= { 0x0c, 0, 0, 0, 0x07 },
  RX_STATUS[]	= { 0x00, 0, 0, 0, 0xE7 },
  MAIN_VFO[]	= { 0x00, 0, 0, 0, 0x03 };

#define	WRITE_RX_VFO	0x01

/* Serial port File descriptor */
static int serial_fd = 0;

/* Original serial port options */
static struct termios serial_old_options;

extern GtkWidget *gbl_popup_menu;

extern pixbuf_data_t wfall_pixbuf;

/*------------------------------------------------------------------------*/

  void
Open_Tcvr_Serial( void )
{
  struct termios new_options; /* New serial port options */
  struct flock lockinfo;      /* File lock information   */

  /* For testing serial port */
  unsigned char test;

  /* Serial port name */
  char serial_port[15];


  /* Abort if serial already open */
  if( isFlagSet(ENABLE_CAT) ) return;

  /* Open Serial device, exit on error */
  ClearFlag( SERIAL_IO_ERROR );
  if( isFlagSet(ENABLE_CAT_847) )
	strcpy( serial_port, rc_data.serial1 );
  else
	if( isFlagSet(ENABLE_CAT_857) )
	  strcpy( serial_port, rc_data.serial2 );

  serial_fd = open( serial_port, O_RDWR | O_NOCTTY );
  if( serial_fd < 0 )
  {
	char message[56];
	SetFlag( SERIAL_IO_ERROR );
	perror( serial_port );
	snprintf( message, 56,
		"Failed to open %s\n"
		"CAT not enabled", serial_port );
	gtk_check_menu_item_set_active( GTK_CHECK_MENU_ITEM(
		  lookup_widget(gbl_popup_menu, "none")), TRUE );
	Error_Dialog( message, FALSE );
	return;
  }

  /* Attempt to lock entire Serial port device file */
  lockinfo.l_type   = F_WRLCK;
  lockinfo.l_whence = SEEK_SET;
  lockinfo.l_start  = 0;
  lockinfo.l_len    = 0;

  /* If Serial device is already locked, abort */
  if( fcntl( serial_fd, F_SETLK, &lockinfo ) < 0 )
  {
	char message[85];
	SetFlag( SERIAL_IO_ERROR );
	perror( serial_port );
	fcntl( serial_fd, F_GETLK, &lockinfo );
	fprintf( stderr, "xfhell: %s: Device locked by pid %d - aborting\n",
		serial_port, lockinfo.l_pid );
	snprintf( message, 85,
		"Failed to lock %s\n"
		"Device locked by pid %d\n"
		"CAT not enabled",
		serial_port, lockinfo.l_pid );
	gtk_check_menu_item_set_active( GTK_CHECK_MENU_ITEM(
		  lookup_widget(gbl_popup_menu, "none")), TRUE );
	Error_Dialog( message, FALSE );
	return;
  }

  /* Save the current serial port options */
  tcgetattr( serial_fd, &serial_old_options );

  /* Read the current serial port options */
  tcgetattr( serial_fd, &new_options );

  /* Set the i/o baud rates */
  if( isFlagSet(ENABLE_CAT_847) )
  {
	cfsetispeed( &new_options, B57600 );
	cfsetospeed( &new_options, B57600 );
  }
  else
	if( isFlagSet(ENABLE_CAT_857) )
	{
	  cfsetispeed( &new_options, B38400 );
	  cfsetospeed( &new_options, B38400 );
	}

  /* Set options for 'raw' I/O mode */
  new_options.c_iflag &=
	~( IGNBRK | BRKINT | PARMRK | ISTRIP | INLCR | IGNCR | ICRNL | IXON );
  new_options.c_oflag &= ~( OPOST | OLCUC | ONLCR | OCRNL | ONOCR | ONLRET );
  new_options.c_lflag &= ~( ECHO | ECHONL | ICANON | ISIG | IEXTEN );
  new_options.c_cflag &= ~( CSIZE | PARENB );
  new_options.c_cflag |= ( CS8 | CLOCAL | CREAD | CSTOPB | HUPCL );

  /* Setup read() timeout to .2 sec */
  new_options.c_cc[ VMIN ]  = 0;
  new_options.c_cc[ VTIME ] = 2;

  /* Setup the new options for the port */
  tcsetattr( serial_fd, TCSAFLUSH, &new_options );

  /* Enable CAT */
  Write_Tcvr_Command(CAT_ON);

  /* Test the serial port */
  Write_Tcvr_Command( RX_STATUS );
  Read_Tcvr_Serial( &test, 1 );

  if( isFlagClear(SERIAL_IO_ERROR) )
	SetFlag( ENABLE_CAT );
  else
  {
	tcsetattr( serial_fd, TCSANOW, &serial_old_options );
	close( serial_fd );
	serial_fd = 0;
	gtk_check_menu_item_set_active( GTK_CHECK_MENU_ITEM(
		  lookup_widget(gbl_popup_menu, "none")), TRUE );
	Error_Dialog( "Failed to Enable CAT", FALSE );	
  }

} /* End of Open_Tcvr_Serial() */

/*-------------------------------------------------------------------------*/

/* Open_RTS_Serial()
 *
 * Opens the Serial Port to be used for RTS control of PTT
 */
  void
Open_RTS_Serial( void )
{
  struct termios new_options; /* New serial port options */
  struct flock lockinfo;      /* File lock information   */

  /* Serial port name */
  char serial_port[15];


  /* Abort if serial already open */
  if( isFlagSet(ENABLE_CAT) ) return;

  /* Open Serial device, exit on error */
  ClearFlag( SERIAL_IO_ERROR );
  strcpy( serial_port, rc_data.serial3 );
  serial_fd = open( serial_port, O_RDWR | O_NOCTTY );
  if( serial_fd < 0 )
  {
	char message[56];
	SetFlag( SERIAL_IO_ERROR );
	perror( serial_port );
	snprintf( message, 56,
		"Failed to open %s\n"
		"RTS Control not enabled", serial_port );
	gtk_check_menu_item_set_active( GTK_CHECK_MENU_ITEM(
		  lookup_widget(gbl_popup_menu, "none")), TRUE );
	Error_Dialog( message, FALSE );
	return;
  }

  /* Attempt to lock entire Serial port device file */
  lockinfo.l_type   = F_WRLCK;
  lockinfo.l_whence = SEEK_SET;
  lockinfo.l_start  = 0;
  lockinfo.l_len    = 0;

  /* If Serial device is already locked, abort */
  if( fcntl( serial_fd, F_SETLK, &lockinfo ) < 0 )
  {
	char message[85];
	SetFlag( SERIAL_IO_ERROR );
	perror( serial_port );
	fcntl( serial_fd, F_GETLK, &lockinfo );
	fprintf( stderr, "xfhell: %s: Device locked by pid %d - aborting\n",
		serial_port, lockinfo.l_pid );
	snprintf( message, 85,
		"Failed to lock %s\n"
		"Device locked by pid %d\n"
		"RTS Control not enabled",
		serial_port, lockinfo.l_pid );
	gtk_check_menu_item_set_active( GTK_CHECK_MENU_ITEM(
		  lookup_widget(gbl_popup_menu, "none")), TRUE );
	Error_Dialog( message, FALSE );
	return;
  }

  /* Save the current serial port options */
  tcgetattr( serial_fd, &serial_old_options );

  /* Read the current serial port options */
  tcgetattr( serial_fd, &new_options );

  /* Set options for RTS control */
  new_options.c_cflag |= ( CRTSCTS | HUPCL );

  /* Setup the new options for the port */
  if( tcsetattr(serial_fd, TCSAFLUSH, &new_options) == -1 )
  {
	tcsetattr( serial_fd, TCSANOW, &serial_old_options );
	close( serial_fd );
	serial_fd = 0;
	gtk_check_menu_item_set_active( GTK_CHECK_MENU_ITEM(
		  lookup_widget(gbl_popup_menu, "none")), TRUE );
	Error_Dialog( "Failed to Enable RTS Control", FALSE );	
  }
  else
	SetFlag( ENABLE_CAT );

} /* End of Open_RTS_Serial() */

/*-------------------------------------------------------------------------*/

/*  Read_Tcvr_Serial()
 *
 *  Reading Data from the Tcvr's Serial Port
 */

  void
Read_Tcvr_Serial( unsigned char *cmnd, int block_size )
{
  /* Error condition */
  if( read(serial_fd, cmnd, block_size) < block_size )
  {
	tcflush( serial_fd, TCIOFLUSH );
	SetFlag( SERIAL_IO_ERROR );
	if( isFlagSet(ENABLE_CAT) )
	  Error_Dialog(
		  "Serial port I/O error\n"
		  "Quit xfhell and correct", TRUE );
  }

} /* End of Read_Tcvr_Serial() */

/*-------------------------------------------------------------------------*/

/*  Read_Rx_Status()
 *
 *  Reads the transceiver's status (Tx/Rx freq, mode,
 *  po/s-meter etc) from the Yaesu transceiver.
 */

  void
Read_Rx_Status( char *mode, char *freq, char *rst )
{
  /* Mask for Rx S-Meter status */
  char FT847_SMETER = 0x1F;
  char FT857_SMETER = 0x0F;

  /* A 5-char string for sending and receiving  */
  /* 5-byte strings to and from the transceiver */
  unsigned char cmnd_parm[5];

  int temp;

  bzero( cmnd_parm, 5 );

  /* Abort if CAT disabled */
  if( isFlagClear(ENABLE_CAT) )
	return;

  /* Read Rx status (S-meter) */
  Write_Tcvr_Command( RX_STATUS );
  if( isFlagSet(SERIAL_IO_ERROR) )
	return;

  Read_Tcvr_Serial( cmnd_parm, 1 );
  if( isFlagSet(SERIAL_IO_ERROR) )
	return;

  /* Mask out S-meter */
  if( isFlagSet(ENABLE_CAT_847) )
  {
	temp = cmnd_parm[0] & FT847_SMETER;
	temp = temp/4 + 3;
  }
  else
	temp = cmnd_parm[0] & FT857_SMETER;

  /* Keep S-meter report in range 3-9 */
  if( temp > 9 )
	temp = 9;
  temp = 500 + 10*temp + 9;
  snprintf( rst, 4, "%3d", temp );

  /* Read Rx mode and frequency */
  Write_Tcvr_Command( MAIN_VFO );
  if( isFlagSet(SERIAL_IO_ERROR) )
	return;

  Read_Tcvr_Serial( cmnd_parm, 5 );
  if( isFlagSet(SERIAL_IO_ERROR) )
	return;

  /* Decode frequency data to 1 kHz */
  temp = 0;
  temp += (cmnd_parm[0] & 0xF0) >> 4;
  temp *= 10;
  temp += cmnd_parm[0] & 0x0F;
  temp *= 10;
  temp += (cmnd_parm[1] & 0xF0) >> 4;
  temp *= 10;
  temp += cmnd_parm[1] & 0x0F;
  temp *= 10;
  temp += (cmnd_parm[2] & 0xF0) >> 4;
  temp *= 10;
  temp += cmnd_parm[2] & 0x0F;


  /* Enter freq data */
  snprintf( freq, 8, "%-7.3f", (float)temp/1000. );

  /* Enter operating mode (LSB/USB etc) */
  *mode = cmnd_parm[4];

} /* End of Read_Rx_Status() */

/*-------------------------------------------------------------------------*/

/*  Write_Tcvr_Command()
 *
 *  Writes a command (status change)
 *  to the Yaesu Tranceiver
 */

  void
Write_Tcvr_Command( unsigned char *cmnd )
{
  tcflush( serial_fd, TCIOFLUSH );
  if( write( serial_fd, cmnd, 5 ) < 5 ) /* Error condition */
  {
	tcflush( serial_fd, TCIOFLUSH );
	SetFlag( SERIAL_IO_ERROR );
	Error_Dialog(
		"Serial port I/O error\n"
		"Quit xfhell and correct", TRUE );
	return;
  }

  /* Give time to CAT to do its thing */
  usleep(1000);

} /* End of Write_Tcvr_Command() */

/*-------------------------------------------------------------------------*/

/*  Set_Tcvr_Mode()
 *
 *  Sets the Transceiver operating mode
 */

  void
Set_Tcvr_Mode( void )
{
  /* For testing serial port */
  unsigned char test;

  /* Set appropriate tcvr mode if CAT enabled */
  if( isFlagSet(ENABLE_CAT) )
  {
	if( isFlagSet(TRANSMIT_MODE) )
	{
	  if( isFlagSet(ENABLE_RTS_CTL) )
		RS232_RTS_Set( TRUE );
	  else
		Write_Tcvr_Command( MODE_USB );
	  if( isFlagSet(SERIAL_IO_ERROR) ) return;

	  if( isFlagSet(ENABLE_CAT_847) )
		Write_Tcvr_Command( PTT_ON );
	  if( isFlagSet(SERIAL_IO_ERROR) ) return;
	}/*  if( isFlagSet(TRANSMIT_MODE) ) */
	else
	{
	  if( isFlagSet(ENABLE_RTS_CTL) )
		RS232_RTS_Set( FALSE );
	  else
		Write_Tcvr_Command( PTT_OFF );
	  if( isFlagSet(SERIAL_IO_ERROR) ) return;

	  if( isFlagSet(ENABLE_CAT_847) )
		Write_Tcvr_Command( MODE_CWN );
	  else
		if( isFlagSet(ENABLE_CAT_857) )
		{
		  sleep(1);
		  Write_Tcvr_Command( MODE_CW );
		}
	  if( isFlagSet(SERIAL_IO_ERROR) ) return;
	} /* else */

	/* Test the serial port */
	if( isFlagClear(ENABLE_RTS_CTL) )
	{	
	  Write_Tcvr_Command( RX_STATUS );
	  Read_Tcvr_Serial( &test, 1 );
	}
  } /* if( isFlagSet(ENABLE_CAT) ) */

} /* Set_Tcvr_Mode() */

/*-------------------------------------------------------------------------*/

/*  Close_Tcvr_Serial()
 *
 *  Restore old options and close the Serial port
 */

  void
Close_Tcvr_Serial( void )
{
  if( isFlagSet(ENABLE_CAT) )
  {
	ClearFlag( SERIAL_IO_ERROR );
	if( isFlagClear( ENABLE_RTS_CTL) )
	  Write_Tcvr_Command( CAT_OFF );
	else
	  RS232_RTS_Set( FALSE );
	ClearFlag( ENABLE_CAT );
	tcsetattr( serial_fd, TCSANOW, &serial_old_options );
	close( serial_fd );
	serial_fd = 0;
  }
} /* End of Close_Tcvr_Serial() */

/*-------------------------------------------------------------------------*/

/* Tune_Tcvr()
 *
 * Tunes the transceiver to the frequency of the strongest
 * signal near a mouse click in the waterfall window
 */

  void
Tune_Tcvr( double x )
{
  int
	from, to,	/* Range to scan for a max bin value  */
	bin_max,	/* Max bin value found in this range  */
	max_idx,	/* fft idx at which the max is found */
	fft_idx,	/* Idx used to search fft bin values */
	tcvr_freq,	/* Transceiver's Rx frequency */
	audio_freq;	/* Audio frequency in waterfal window */

  /* Calculate fft index corresponding to pointer x in waterfall window */
  fft_idx = ( ((int)(x-0.5))*FFT_SIZE)/(2*(wfall_pixbuf.width-2) );

  /* Look above and below click point for max bin val */
  from = fft_idx - 10;
  if( from < 0 )
	from = 0;
  to = fft_idx + 10;
  if( to >= FFT_SIZE/2 )
	to = FFT_SIZE/2;

  /* Find max bin value around click point */
  bin_max = 0;
  max_idx = fft_idx;
  for( fft_idx = from; fft_idx < to; fft_idx++ )
	if( bin_max < gbl_bin_ave[fft_idx] )
	{
	  bin_max = gbl_bin_ave[fft_idx];
	  max_idx = fft_idx;
	}

  /* Audio freq. corresponding to fft index */
  audio_freq = (max_idx * 4 * rc_data.tone_freq) / FFT_SIZE;

  /* Read current Rx frequency */
  tcvr_freq = Read_Rx_Freq();

  /* Calculate and write Rx freq. to center audio on rc_data.tone_freq */
  Write_Rx_Freq(tcvr_freq + audio_freq - rc_data.tone_freq);

} /* Tune_Tcvr() */

/*-------------------------------------------------------------------------*/

/*  Read_Rx_Freq()
 *
 *  Reads the Rx freq of the Yaesu transceiver.
 */

  int
Read_Rx_Freq( void )
{
  /* A 5-char string for sending and receiving  */
  /* 5-byte strings to and from the transceiver */
  unsigned char cmnd_parm[5];

  int freq;

  bzero( cmnd_parm, 5 );

  /* Abort if CAT disabled */
  if( isFlagClear(ENABLE_CAT) ) return(-1);

  /* Read Rx frequency */
  Write_Tcvr_Command( MAIN_VFO );
  if( isFlagSet(SERIAL_IO_ERROR) ) return(-2);

  Read_Tcvr_Serial( cmnd_parm, 5 );
  if( isFlagSet(SERIAL_IO_ERROR) ) return(-3);

  /* Decode frequency data to 10Hz */
  freq = 0;
  freq += (cmnd_parm[0] & 0xF0) >> 4;
  freq *= 10;
  freq += cmnd_parm[0] & 0x0F;
  freq *= 10;
  freq += (cmnd_parm[1] & 0xF0) >> 4;
  freq *= 10;
  freq += cmnd_parm[1] & 0x0F;
  freq *= 10;
  freq += (cmnd_parm[2] & 0xF0) >> 4;
  freq *= 10;
  freq += cmnd_parm[2] & 0x0F;
  freq *= 10;
  freq += (cmnd_parm[3] & 0xF0) >> 4;
  freq *= 10;
  freq += cmnd_parm[3] & 0x0F;
  freq *= 10;
  freq += (cmnd_parm[4] & 0xF0) >> 4;

  return( freq );

} /* End of Read_Rx_Freq() */

/*------------------------------------------------------------------------*/

/*  Write_Rx_Freq()
 *
 *  Writes the Rx freq to the Yaesu FT-847 transceiver.
 */

  void
Write_Rx_Freq( int freq )
{
  /* A 5-char string for sending and receiving */
  /* 5-byte strings to and from the transceiver */
  unsigned char cmnd_parm[5];

  /* Buffer used for converting freq. to string */
  char freq_buff[9];

  int idx; /* Index for loops etc */

  bzero( cmnd_parm, 5 );

  /* Set Rx frequency */
  snprintf( freq_buff, 9, "%08d", freq/10 );
  for( idx = 0; idx < 4; idx++ )
  {
	cmnd_parm[idx]  = (freq_buff[2*idx]   - '0') << 4;
	cmnd_parm[idx] |= (freq_buff[2*idx+1] - '0');
  }
  cmnd_parm[4] = WRITE_RX_VFO;
  Write_Tcvr_Command( cmnd_parm );

} /* End of Write_Rx_Freq() */

/*-------------------------------------------------------------------------*/

/* RS232_RTS_Set()
 *
 * Sets the Serial Port RTS line UP/DOWN
 */
  void
RS232_RTS_Set( gboolean set )
{
  int modem_lines;

  /* Read current modem state */
  if( ioctl(serial_fd, TIOCMGET, &modem_lines) == -1 )
  {
	SetFlag( SERIAL_IO_ERROR );
	perror( rc_data.serial3 );
	gtk_check_menu_item_set_active( GTK_CHECK_MENU_ITEM(
		  lookup_widget(gbl_popup_menu, "none")), TRUE );
	Error_Dialog(
		"Failed to set RTS\n"
		"RTS Control disabled", FALSE );
	return;
  }

  /* Change the RTS state */
  if( set )
	modem_lines |=  (int)TIOCM_RTS;
  else
	modem_lines &= ~(int)TIOCM_RTS;

  /* Write the new setting */
  if( ioctl(serial_fd, TIOCMSET, &modem_lines) == -1 )
  {
	SetFlag( SERIAL_IO_ERROR );
	perror( rc_data.serial3 );
	gtk_check_menu_item_set_active( GTK_CHECK_MENU_ITEM(
		  lookup_widget(gbl_popup_menu, "none")), TRUE );
	Error_Dialog(
		"Failed to set RTS\n"
		"RTS Control disabled", FALSE );
	return;
  }

  return;
} /* RS232_RTS_Set() */

/*-------------------------------------------------------------------------*/

