/*  codec.c
 *
 *  Feld Hell coder/decoder functions of xfhell application
 */

/*
 *  xfhell: An application to transmit and receive
 *  Feld Hell signals using a computer's sound card
 *
 *  Copyright (C) 2003  Neoklis Kyriazis
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of
 *  the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details:
 *
 *  http://www.gnu.org/copyleft/gpl.txt
 */

#include "xfhell.h"
#include <math.h>

extern int
  contrast, /* Contrast value for Receive window */
  resample; /* Resample request in DSP samples   */

/* Signal scope */
extern GtkWidget *gbl_scope;

/* Signal/DSP samples buffer. 2*size for Stereo */
extern unsigned char *dsp_buff;
int tx_buff_idx;

/* Runtime configuration data */
extern rc_data_t rc_data;

/* sine and cosine lookup tables */
int
  *sine_table   = NULL,
  *cosine_table = NULL;

/*------------------------------------------------------------------------*/

/*  Assemble_Column()
 *
 *  Assembles a character column ( 14 or 12 dots) from incoming
 *  Hellschreiber pulse train. The column is stored as an array
 *  of signal intensity values for the elements that comprise
 *  each dot, depending on dot size. For a 2x4 pixel/dot font size,
 *  there would be 4 elements/dot e.g. 1 elements/pixel vertically.
 */

  gboolean
Assemble_Column( unsigned char *column )
{
  int
	elem[rc_data.bitmap_hgt], 	/* Element values for a column */
	elem_lev, 	/* Signal level during a element  */
	elem_max,	/* Max value of column's elements */
	row_min,	/* Minimum value of row averages  */
	row_max,	/* Maximum value of row averages  */
	col_ave,	/* Average value of row averages  */
	min_idx,	/* Index to row with minimum value */
	max_idx,	/* Index to row with maximum value */
	temp,
	idx;

  /* Average value of element max */
  static int
	*row_ave = NULL, 	/* Average element value per row */
	fbbh = 0,			/* Font bbx height flag */
	resample_intg = 0,	/* Integral of resample request  */
	resample_idx = 0,	/* Index used to calc resampling */
	ave_max = 0;		/* Average of element max values */

  /* Allocate memory to row_ave */
  if( fbbh != rc_data.font_data.fbb_h )
  {
	fbbh = rc_data.font_data.fbb_h;
	row_ave = realloc( row_ave, sizeof(int)*(size_t)fbbh );
	if( row_ave == NULL )
	{
	  perror( "row_ave" );
	  Error_Dialog(
		  "realloc() for row_ave failed\n"
		  "Quit xfhell and correct", TRUE );
	  ClearFlag( RECEIVE_MODE );
	  ClearFlag(
		  TRANSMIT_MODE  |
		  TRANSMIT_MACRO |
		  TRANSMIT_TAG   |
		  TRANSMIT_KEYBD );

	  return( FALSE );
	}
	bzero( row_ave, sizeof(int)*(size_t)fbbh );
  }

  /* Assemble a column's dots. The entries in the elem */
  /* array represent "elements" of dots. There is one  */
  /* element/pixel in a dot, according to its height.  */
  elem_max = min_idx = max_idx = row_max = col_ave = 0;
  row_min = 100000;
  for( idx = 0; idx < rc_data.bitmap_hgt; idx++ )
  {
	if( (elem_lev = Get_Element()) == -1 )
	  return( FALSE );
	elem[idx] = elem_lev;

	/* Find average value of each row of a character.
	 * Needed for automatic deskewing of the display. */
	if( isFlagSet(AUTO_DESKEW) )
	{
	  temp = (2*idx)/rc_data.dot_hgt;
	  row_ave[temp] =
		(row_ave[temp]*(ROW_AVE_WIN-1) + elem_lev*ROW_AVE_WIN)/ROW_AVE_WIN;
	  col_ave += row_ave[temp];

	  /* Find row with the lowest value */
	  if( row_ave[temp] < row_min )
	  {
		row_min = row_ave[temp];
		min_idx = temp;
	  }

	  /* Find row with the highest value */
	  if( row_ave[temp] > row_max )
	  {
		row_max = row_ave[temp];
		max_idx = temp;
	  }

	} /* if( isFlagSet(AUTO_DESKEW) ) */

	/* Find max of element levels */
	if( elem_lev > elem_max )
	  elem_max = elem_lev;

  } /* for( idx = 0; idx < rc_data.bitmap_hgt; idx++ ) */

  /* Try to correct skew by adjusting sampling freq */
  if( isFlagSet(AUTO_DESKEW) )
  {
	gboolean test;
	/* Select either index to min level row (usually
	 * when characters are being sent) or index to max
	 * level row (usually when idle dots are being sent)
	 * On reverse video the test has to be reversed */
	col_ave /= rc_data.bitmap_hgt;
	test = row_max > (3 * col_ave);
	if( rc_data.rev_video ) test = !test;
	if( test ) /* Generally receiving characters (normal video) */
	  resample_idx = max_idx;
	else /* Generally receiving idle dots (normal video) */
	  resample_idx = min_idx;

	/* Keep index in range */
	if( resample_idx >= rc_data.font_data.fbb_h >> 1 )
	  resample = rc_data.font_data.fbb_h - resample_idx;
	else
	  resample = -resample_idx;

	/* Integrate error if in limits */
	if( (resample_intg <  256 && resample > 0) ||
		(resample_intg > -256 && resample < 0) )
	  resample_intg += resample;

	/* Accelerate correction for high skews */
	resample *= abs( resample );
	/* Add integral term */
	resample += resample_intg/32;
  }

  /* Sliding window average of element max */
  ave_max =
	(ave_max*(MAXSEG_AVE_WIN-1) + elem_max*MAXSEG_AVE_WIN)/MAXSEG_AVE_WIN;

  /*** Scale element values to a max of 255 ***/
  /* Limits up-scaling of noise */
  temp = contrast * MAXSEG_AVE_WIN;
  if( ave_max < temp ) ave_max = temp;
  if( !ave_max ) ave_max = 1;

  for( idx = 0; idx < rc_data.bitmap_hgt; idx++ )
  {
	if( ++min_idx >= rc_data.bitmap_hgt )
	  min_idx = 0;

	elem_lev = 255 * MAXSEG_AVE_WIN * elem[idx] / ave_max;
	if( elem_lev > 255 )
	  column[idx] = 255;
	else
	  column[idx] = elem_lev;
  }

  return( TRUE );

} /* Assemble_Column() */

/*------------------------------------------------------------------------*/

/* Transmit_Character()
 *
 * Transmits one Hell character pointed by "glx"
 */

  gboolean
Transmit_Character( font_data_t *font_data, int glx )
{
  if( isFlagSet(MODE_FELDHELL) )
	return( Tx_Feld_Character(font_data, glx) );
  else
	return( Tx_FMHell_Character(font_data, glx) );
} /* Transmit_Character() */

/*------------------------------------------------------------------------*/

/* Tx_FMHell_Character()
 *
 * Transmits one FM Hell character pointed by "glx"
 */

  gboolean
Tx_FMHell_Character( font_data_t *font_data, int glx )
{
  int
	buff_idx, /* Index to audio samples buffer */
	bmx,	  /* Bitmap x index */
	bmy,	  /* Bitmap y index */
	dot,      /* State of dot to transmit  */
	idx;

  static double
	twopi = 2.0 * M_PI,
	tone_phase,		/* Phase of audio tone */
	white_phinc,	/* Phase increment for white tone */
	black_phinc,	/* Phase increment for black tone */
	dot_resol = 0,
	baud_rate = 0.0;

  /* Column buffer for drawing chars */
  unsigned char column[rc_data.bitmap_hgt];

  /* Initialize on new parameters or first call */
  if( (baud_rate != rc_data.baud_rate) ||
	  (dot_resol != rc_data.dot_resol) )
  {
	baud_rate = rc_data.baud_rate;
	dot_resol = rc_data.dot_resol;

	white_phinc = twopi /
	   (double)rc_data.dsp_speed *
	  ((double)rc_data.tone_freq + baud_rate *
	   (double)rc_data.dot_resol / 4.0);
	black_phinc = twopi /
	   (double)rc_data.dsp_speed *
	  ((double)rc_data.tone_freq - baud_rate *
	   (double)rc_data.dot_resol / 4.0);

	tone_phase  = 0.0;
  }

  /* Adjust glyph index */
  glx -= font_data->first_glyph;

  /* Transmit columns */
  for( bmx = 0; bmx < font_data->glyph_data[glx].dwidth_x; bmx++ )
  {
	/* Initialize buffer idx according to stereo/mono mode */
	if( rc_data.num_chan == 2 )
	  buff_idx = rc_data.chan_used;
	else
	  buff_idx = 0;

	/* Go up in bitmap buffer row by row */
	for( bmy = font_data->fbb_h-1; bmy >= 0; bmy-- )
	{
	  /* Set tone control according to bitmap (dot) value */
	  dot = font_data->glyph_data[glx].bitmap[bmy] & ( 1 << (31-bmx) );
	  /* Reverse video if selected */
	  if( rc_data.rev_video ) dot = !dot;
	  if( dot ) /* Black dot, send black tone */
	  {
		/* Fill buffer with samples at black tone rate */
		for( idx = 0; idx < rc_data.samples_perdot2; idx++ )
		{
		  dsp_buff[buff_idx] = (unsigned char)(127.0 * sin(tone_phase)) + 0x80;
		  buff_idx   += rc_data.num_chan;
		  tone_phase += black_phinc;
		  if( tone_phase >= twopi ) tone_phase -= twopi;
		}
	  } /* if( dot ) */
	  else /* Blank dot */
		/* Fill buffer with samples at white tone rate */
		for( idx = 0; idx < rc_data.samples_perdot2; idx++ )
		{
		  dsp_buff[buff_idx] = (unsigned char)(127.0 * sin(tone_phase)) + 0x80;
		  buff_idx   += rc_data.num_chan;
		  tone_phase += white_phinc;
		  if( tone_phase >= twopi ) tone_phase -= twopi;
		}

	} /* for( bmy = temp; bmy >= 0; bmy-- ) */

	/*** Write samples buffer to DSP ***/
	DSP_Write( dsp_buff, rc_data.txbuff_len );
	if( isFlagSet(DSP_IO_ERROR) )
	  return( FALSE );

	/* Loopback Tx to Rx */
	if( isFlagSet(TX2RX_LOOPBACK) )
	{
	  /* Reset tx buffer idx */
	  if( rc_data.num_chan == 2 )
		tx_buff_idx = rc_data.chan_used;
	  else
		tx_buff_idx = 0;

	  /* Draw characters as they are xmitted */
	  if( ! Assemble_Column(column) )
		return( FALSE );
	  if( ! Draw_Column(column) )
		return( FALSE );

	} /* if( isFlagSet(TX2RX_LOOPBACK) ) */

  } /* for( bmx = 0; bmx < font_data->glyph_data[glx].dwidth_x; bmx++ ) */

  return( TRUE );

} /* Tx_FMHell_Character() */

/*------------------------------------------------------------------------*/

/* Tx_Feld_Character()
 *
 * Transmits one Feld Hell character pointed by "glx"
 */

  gboolean
Tx_Feld_Character( font_data_t *font_data, int glx )
{
  int
	buff_idx,	/* Index to audio samples buffer */
	cosine_idx,	/* Index to cos wave lookup table */
	bmx,		/* Bitmap x index */
	bmy,		/* Bitmap y index */
	idb, idx;

  static int
	tone_ctrl = 0, /* Tone raise/lower control indx */
	sine_idx  = 0; /* Index to sine wave lookup table */

  int
	last_dot, /* State of last dot transmitted */
	new_dot;  /* State of new dot to transmit  */

  /* Column buffer for drawing chars */
  unsigned char column[rc_data.bitmap_hgt];


  /* Adjust glyph index */
  glx -= font_data->first_glyph;

  /* Transmit columns */
  bmy = font_data->fbb_h-1;
  for( bmx = 0; bmx < font_data->glyph_data[glx].dwidth_x; bmx++ )
  {
	/* Initialize buffer idx according to stereo/mono mode */
	if( rc_data.num_chan == 2 )
	  buff_idx = rc_data.chan_used;
	else
	  buff_idx = 0;

	/* Go up in bitmap buffer row by row */
	for( idb = font_data->fbb_h-1; idb >= 0; idb-- )
	{
	  /* Set tone control according to bitmap value.  */
	  if( --bmy < 0 ) bmy = font_data->fbb_h-1;
	  new_dot  =
		font_data->glyph_data[glx].bitmap[bmy] & ( 1 << (31-bmx) );
	  last_dot =
		font_data->glyph_data[glx].bitmap[idb] & ( 1 << (31-bmx) );

	  /* Reverse video if activated */
	  if( rc_data.rev_video )
	  {
		new_dot  = !new_dot;
		last_dot = !last_dot;
	  }

	  /* If two consecutive 1's raise tone, else drop */
	  if( new_dot && last_dot )
		tone_ctrl |= RAISE_TONE;
	  else
		tone_ctrl &= LOWER_TONE;

	  /* Raise/Lower audio tone as requested */
	  switch( tone_ctrl )
	  {
		case 0: /*** No tone (carrier down) ***/

		  /* Fill buffer with zero level samples */
		  for( idx = 0; idx < rc_data.samples_perdot2; idx++ )
		  {
			dsp_buff[buff_idx] = 0x80;
			buff_idx += rc_data.num_chan;
			sine_idx++;
			if( sine_idx >= rc_data.samples_percycle )
			  sine_idx = 0;
		  }

		  break;

		case 1: /*** Raise tone (carrier up) ***/

		  /* Fill buffer with cosine shaped rising edge */
		  for( cosine_idx = 0; cosine_idx < rc_data.samples_perdot2; cosine_idx++ )
		  {
			dsp_buff[buff_idx] = (unsigned char)
			  ((cosine_table[cosine_idx] * sine_table[sine_idx])>>24) + 0x80;
			buff_idx += rc_data.num_chan;
			sine_idx++;
			if( sine_idx >= rc_data.samples_percycle )
			  sine_idx = 0;
		  }
		  tone_ctrl |= TONE_RAISED;

		  break;

		case 2: /*** Lower tone (Carrier down) ***/

		  /* Fill buffer with cosine shaped falling edge */
		  for( cosine_idx = rc_data.samples_perdot2-1; cosine_idx >= 0; cosine_idx-- )
		  {
			dsp_buff[buff_idx] = (unsigned char)
			  ((cosine_table[cosine_idx] * sine_table[sine_idx]) >> 24) + 0x80;
			buff_idx += rc_data.num_chan;
			sine_idx++;
			if( sine_idx >= rc_data.samples_percycle )
			  sine_idx = 0;
		  }
		  tone_ctrl &= TONE_LOWERED;

		  break;

		case 3: /*** Keep tone raised (carrier up) ***/

		  /* Fill buffer with steady tone */
		  for( idx = 0; idx < rc_data.samples_perdot2; idx++ )
		  {
			dsp_buff[buff_idx] = (sine_table[sine_idx] >> 8) + 0x80;
			buff_idx += rc_data.num_chan;
			sine_idx++;
			if( sine_idx >= rc_data.samples_percycle )
			  sine_idx = 0;
		  }

	  } /* switch( tone_ctrl ) */

	} /* for( bmy = font_data->fbb_h-1; bmy >= 0; bmy-- ) */

	/*** Write samples buffer to DSP ***/
	DSP_Write( dsp_buff, rc_data.txbuff_len );
	if( isFlagSet(DSP_IO_ERROR) )
	  return( FALSE );

	/* Loopback Tx to Rx */
	if( isFlagSet(TX2RX_LOOPBACK) )
	{
	  /* Reset tx buffer idx */
	  if( rc_data.num_chan == 2 )
		tx_buff_idx = rc_data.chan_used;
	  else
		tx_buff_idx = 0;

	  /* Draw characters as they are xmitted */
	  if( ! Assemble_Column(column) )
		return( FALSE );
	  if( ! Draw_Column(column) )
		return( FALSE );

	} /* if( isFlagSet(TX2RX_LOOPBACK) ) */

  } /* for( bmx = 0; bmx < font_data->glyph_data[glx].dwidth_x; bmx++ ) */

  return( TRUE );

} /* Tx_Feld_Character() */

/*------------------------------------------------------------------------*/

/* Morse_Transmit()
 *
 * Transmits a string Morse code
 */

  void
Morse_Transmit( char *mesg )
{
  /* DSP samples buffer.
   * Size is (samples/cycle)*(audio freq)*(dash duration)
   * Dash duration is abt 3/wpm seconds.
   * In stereo mode the buffer size must be doubled
   */
  unsigned char
	samples_buff[ 3*rc_data.num_chan*rc_data.dsp_speed/rc_data.wpm ];

  int
	mesg_len,   /* String length of message to send  */
	mesg_idx,   /* Index into the message string     */
	morse_chr,  /* Morse code character in hex equiv */
	morse_idx,  /* Index into the Morse Hex code     */
	elem_dur,   /* Duration in samples of Morse elem */
	dash_dur,   /* Duration in samples of Morse dash */
	tone_dur,   /* Duration in samples of trans tone */
	buff_idx,   /* Index to audio samples buffer     */
	buff_size,  /* Buffer size to be written to DSP  */
	sine_idx,   /* Index to sine wave lookup table   */
	cosine_idx, /* Index to cosine wave lookup table */
	idx;

  /* Tables of Morse and ascii characters */
  int
	ascii_char[] = ASCII_CHAR,
	morse_code[] = MORSE_CODE;


  /* Duration in samples of Morse element */
  elem_dur = rc_data.dsp_speed / rc_data.wpm;
  dash_dur = 3 * elem_dur;
  buff_size = dash_dur * rc_data.num_chan;

  /* Clean up buffer */
  for( buff_idx = 0; buff_idx < buff_size; buff_idx++ )
	samples_buff[buff_idx] = 0x80;

  /* Convert ascii characters to Morse hex code */
  mesg_len = strlen( mesg );
  for( mesg_idx = 0; mesg_idx < mesg_len; mesg_idx++ )
  {
	/* Find equivalent Hex morse code */
	for( idx = 0; idx < NUM_MORSE_CHARS; idx++ )
	  if( mesg[mesg_idx] == ascii_char[idx] )
		break;
	morse_chr = morse_code[idx];

	/* Find beginning of Morse code (1st '1') */
	for( idx = 8; idx >= 0; idx-- )
	  if( morse_chr & (1 << idx) )
		break;

	morse_idx = idx - 1;
	/* Transmit dots (1's) or dashes (0's) */
	while( morse_idx >= 0 )
	{
	  /* If element is dot ('1') */
	  if( morse_chr & (1 << morse_idx) )
		/* Tone duration in samples - edges */
		tone_dur = elem_dur - rc_data.samples_perdot;
	  else /* If element is dash ('0') */
		/* Tone duration in samples is triple */
		tone_dur = dash_dur - rc_data.samples_perdot;

	  /* Total samples to be generated */
	  buff_size = (tone_dur + rc_data.samples_perdot) * rc_data.num_chan;

	  /* Initialize buffer idx according to stereo/mono mode */
	  if( rc_data.num_chan == 2 )
		buff_idx = rc_data.chan_used;
	  else
		buff_idx = 0;

	  /* Fill buffer with cosine shaped rising edge */
	  sine_idx = 0;
	  for( cosine_idx = 0; cosine_idx < rc_data.samples_perdot2; cosine_idx++ )
	  {
		samples_buff[buff_idx] =
		  ((cosine_table[cosine_idx] * sine_table[sine_idx])>>24) + 0x80;
		buff_idx += rc_data.num_chan;
		if( ++sine_idx == rc_data.samples_percycle )
		  sine_idx = 0;
	  }

	  /* Fill buffer with the steady tone samples */
	  for( idx = 0; idx < tone_dur; idx++ )
	  {
		samples_buff[buff_idx] = (sine_table[sine_idx]>>8) + 0x80;
		buff_idx += rc_data.num_chan;
		if( ++sine_idx == rc_data.samples_percycle )
		  sine_idx = 0;
	  }

	  /* Fill buffer with cosine shaped falling edge */
	  for( cosine_idx = rc_data.samples_perdot2-1; cosine_idx >= 0; cosine_idx-- )
	  {
		samples_buff[buff_idx] =
		  ((cosine_table[cosine_idx] * sine_table[sine_idx])>>24) + 0x80;
		buff_idx += rc_data.num_chan;
		if( ++sine_idx == rc_data.samples_percycle )
		  sine_idx = 0;
	  }

	  /*** Write samples buffer to DSP, abort on error ***/
	  DSP_Write( samples_buff, buff_size );
	  if( isFlagSet(DSP_IO_ERROR) )
		return;

	  /* Initialize buffer according to stereo/mono mode */
	  if( rc_data.num_chan == 2 )
		buff_idx = rc_data.chan_used;
	  else
		buff_idx = 0;

	  /* Fill buffer with 0 level for space */
	  /* If inter-word space add 3 extra space elements */
	  if( morse_chr == 0x01 )
	  {
		for( idx = 0; idx < dash_dur; idx++ )
		{
		  samples_buff[buff_idx] = 0x80;
		  buff_idx += rc_data.num_chan;
		}
		buff_size = dash_dur * rc_data.num_chan;

		/*** Writebuffer to DSP, abort on error ***/
		DSP_Write( samples_buff, buff_size );
		if( isFlagSet(DSP_IO_ERROR) )
		  return;

	  }/*  if( morse_chr == 0x01 ) */

	  /* Add inter-element space */
	  for( idx = 0; idx < elem_dur; idx++ )
	  {
		samples_buff[buff_idx] = 0x80;
		buff_idx += rc_data.num_chan;
	  }
	  buff_size = elem_dur * rc_data.num_chan;

	  /*** Write buffer to DSP, abort on error ***/
	  DSP_Write( samples_buff, buff_size );
	  if( isFlagSet(DSP_IO_ERROR) )
		return;

	  morse_idx--;

	} /* while( morse_idx >= 0 ) */

	/* Send inter-character space */
	/* Initialize buffer according to stereo/mono mode */
	if( rc_data.num_chan == 2 )
	  buff_idx = rc_data.chan_used;
	else
	  buff_idx = 0;

	for( idx = 0; idx < 2 * elem_dur; idx++ )
	{
	  samples_buff[buff_idx] = 0x80;
	  buff_idx += rc_data.num_chan;
	}

	buff_size = 2 * elem_dur * rc_data.num_chan;

	/*** Write samples buffer to DSP ***/
	DSP_Write( samples_buff, buff_size );
	if( isFlagSet(DSP_IO_ERROR) )
	  return;

  } /* for( mesg_idx = 0; mesg_idx < mesg_len; mesg_idx++ ) */

} /* Morse_Transmit() */

/*------------------------------------------------------------------------*/

/* Make_Sin_Wavetable()
 *
 * Makes a sine look-up wavetable
 */

  void
Make_Sin_Wavetable()
{
  double sine, angle;
  int idx;

  /* Allocate buffer */
  sine_table = realloc( sine_table, sizeof(int)*(size_t)rc_data.samples_percycle );
  if( sine_table == NULL )
  {
	perror( "sine_table" );
	Error_Dialog(
		"malloc() for sine_table failed\n"
		"Quit xfhell and correct", TRUE );
	return;
  }

  for( idx = 0; idx < rc_data.samples_percycle; idx++ )
  {
	angle = 2.0*M_PI*
	  (double)idx / (double)rc_data.samples_percycle;
	sine = sin( angle );
	sine_table[ idx ] = (int) (sine * 32767.0);
  }

} /* Make_Sin_Wavetable() */

/*------------------------------------------------------------------------*/

/* Make_Cos_Wavetable()
 *
 * Makes a cosine look-up wavetable
 */

  void
Make_Cos_Wavetable()
{
  double angle, cosine;
  int idx;

  /* Allocate buffer */
  cosine_table = realloc( cosine_table, sizeof(int)*(size_t)rc_data.samples_perdot2 );
  if( cosine_table == NULL )
  {
	perror( "cosine_table" );
	Error_Dialog(
		"realloc() for cosine_table failed\n"
		"Quit xfhell and correct", TRUE );
	return;
  }

  for( idx = 0; idx < rc_data.samples_perdot2; idx++ )
  {
	angle = M_PI_2*
	  (double)(rc_data.samples_perdot2-idx)/(double)rc_data.samples_perdot2;
	cosine = cos( angle );
	cosine_table[ idx ] = (int) (cosine * 65536.0);
  }

} /* Make_Cos_Wavetable() */

/*------------------------------------------------------------------------*/

