/*  display.c
 *
 *  Display/Rendering functions for xfhell
 */

/*
 *  xfhell: An application to transmit and receive
 *  Feld Hell signals using a computer's sound card
 *
 *  Copyright (C) 2003  Neoklis Kyriazis
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of
 *  the License, or (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details:
 *
 *  http://www.gnu.org/copyleft/gpl.txt
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gtk/gtk.h>
#include "xfhell.h"
#include "support.h"

/* xfhell main window */
extern GtkWidget *main_window;

/* Global widgets */
extern GtkWidget
  *gbl_scope,
  *gbl_drawingarea,
  *gbl_tx_scrolledwindow;

/* Receive window and waterfall pixbuffs */
pixbuf_data_t
  wfall_pixbuf,
  receive_pixbuf;

/* Text buffers for text viewers */
extern GtkTextBuffer *gbl_tx_text_buffer;

/* Index to macros */
int gbl_macro_idx;

/* Runtime config data */
extern rc_data_t rc_data;

/* Average bin values */
int gbl_bin_ave[FFT_SIZE/2];

/* fft in/out (real/imaginary) buffers */
extern int
  *fft_in_r,
  *fft_out_r,
  *fft_out_i;

/*------------------------------------------------------------------------*/

/*  Draw_Column()
 *
 *  Draws a character's column in the Rx drawing area
 */

  gboolean
Draw_Column( unsigned char *column )
{
  /* Index to current pixel in pixbuf */
  static guchar	*chr_pix;

  /* Char row height, drawing area column and row */
  static int
	scrn_fill,	/* How much of Rx screen is filled */
	row_hgt,	/* Total height of rows on screen */
	chr_col=0,	/* Current column when painting characters */
	chr_row;	/* Current row when painting characters */

  int
	idx, i, j, temp,
	pix_idx; /* Index to pixel in column */

  static unsigned char *last_column = NULL;


  /* Clear Receive window */
  if( isFlagSet(CLEAR_RX_WINDOW) )
  {
	ClearFlag( CLEAR_RX_WINDOW );
	if( chr_col > 2 ) Save_Pixbuf();
	Clear_Pixbuf( &receive_pixbuf, gbl_drawingarea );
	row_hgt = rc_data.num_rows * rc_data.bitmap_hgt;
	chr_col = 2;
	scrn_fill = 0;

	/* Allocate/clear memory for column buffer */
	last_column = realloc( last_column,
		sizeof(unsigned char) * (size_t)rc_data.bitmap_hgt );
	if( last_column == NULL )
	{
	  perror( "last_column" );
	  Error_Dialog(
		  "realloc() for last_column failed\n"
		  "Quit xfhell and correct", TRUE );
	  ClearFlag( RECEIVE_MODE );
	  ClearFlag(
		  TRANSMIT_MODE  |
		  TRANSMIT_MACRO |
		  TRANSMIT_TAG   |
		  TRANSMIT_KEYBD );
	  return( FALSE );
	}
	memset( last_column, 0,
		sizeof(unsigned char) * (size_t)rc_data.bitmap_hgt );

	/* Put cursor and char line at bottom of display */
	chr_row = gbl_drawingarea->allocation.height-2 - row_hgt;
	chr_pix = receive_pixbuf.pixels +
	  (gbl_drawingarea->allocation.height-2 - row_hgt) *
	  receive_pixbuf.rowstride + 2*receive_pixbuf.n_channels;
  } /* if( isFlagSet(CLEAR_RX_WINDOW) ) */

  /*** Draw a character column ***/
  /* Number of columns depends on dot size: */
  /* 1 for 1x2, and 2 and 2x4 pixels */
  for( idx = 0; idx < rc_data.dot_wdt; idx++ )
  {
	/* Draw elements, one per pixel, to pixbuf */
	for( pix_idx = 0; pix_idx < rc_data.bitmap_hgt; pix_idx++ )
	{
	  unsigned char grlev; /* dot gray level */

	  i = receive_pixbuf.rowstride * (rc_data.bitmap_hgt - pix_idx - 1);
	  grlev = 0xff - column[pix_idx];
	  chr_pix[i  ] = grlev;
	  chr_pix[i+1] = grlev;
	  chr_pix[i+2] = grlev;
	}

	/* Draw 2nd raw if traditional 2-row mode is in use */
	if( isFlagClear(AUTO_DESKEW) )
	{
	  for( pix_idx = 0; pix_idx < rc_data.bitmap_hgt; pix_idx++ )
	  {
		unsigned char grlev; /* dot gray level */

		i = receive_pixbuf.rowstride * (row_hgt - pix_idx - 1);
		grlev = 0xff - last_column[pix_idx];
		chr_pix[i  ] = grlev;
		chr_pix[i+1] = grlev;
		chr_pix[i+2] = grlev;
	  }
	}

	/* Draw new column */
	gtk_widget_queue_draw_area( gbl_drawingarea, chr_col, chr_row, 1, row_hgt );

	chr_col++;
	chr_pix += receive_pixbuf.n_channels;

	/* Move back to beginning of row when reaching end of row */
	if( chr_col >= receive_pixbuf.width-2 )
	{
	  chr_col = rc_data.bitmap_wdt + 2;
	  chr_pix = receive_pixbuf.pixels +
		chr_col * receive_pixbuf.n_channels +
		chr_row * receive_pixbuf.rowstride;

	  /* Save pixbuf when screen filled */
	  scrn_fill += row_hgt;
	  if( scrn_fill >= gbl_drawingarea->allocation.height-4 )
	  {
		Save_Pixbuf();
		scrn_fill = 0;
	  }

	  /* Scroll up display one row */
	  gdk_pixbuf_copy_area(
		  receive_pixbuf.pixbuf,
		  2, row_hgt+2,
		  gbl_drawingarea->allocation.width-4,
		  gbl_drawingarea->allocation.height-4 - row_hgt,
		  receive_pixbuf.pixbuf, 2, 2 );

	  /* Redraw last block to beginning of new line */
	  gdk_pixbuf_copy_area(
		  receive_pixbuf.pixbuf,
		  receive_pixbuf.width-2 - rc_data.bitmap_wdt, chr_row,
		  rc_data.bitmap_wdt, row_hgt,
		  receive_pixbuf.pixbuf, 2, chr_row );

	  /* Clear new row */
	  for( pix_idx = 0; pix_idx < row_hgt; pix_idx++ )
	  {
		i = (gbl_drawingarea->allocation.height-3 - pix_idx) *
		  receive_pixbuf.rowstride;
		i += chr_col * receive_pixbuf.n_channels;
		temp = gbl_drawingarea->allocation.width-3;
		for( j = chr_col; j < temp; j++ )
		{
		  i += receive_pixbuf.n_channels;
		  receive_pixbuf.pixels[i  ] = 0xff;
		  receive_pixbuf.pixels[i+1] = 0xff;
		  receive_pixbuf.pixels[i+2] = 0xff;
		}
	  }

	  gtk_widget_queue_draw( gbl_drawingarea );

	} /* if( chr_col >= gbl_width-2 ) */

  } /* for( idx = 0; idx < rc_data.dot_wdt; idx++ ) */

  memcpy( last_column, column, (size_t)rc_data.bitmap_hgt );
  return( TRUE );

} /* Draw_Column() */

/*------------------------------------------------------------------------*/

/* Display_Waterfall()
 *
 * Displays audio spectrum as "waterfall"
 */

  void
Display_Waterfall( void )
{
  int
	idh, idv,  /* Index to hor. and vert. position in warterfall */
	pixel_val, /* Greyscale value of pixel derived from fft o/p  */
	fft_idx;   /* Index to fft output array */

  int
	bin_val,   /* Value of fft output "bin" */
	bin_max=0, /* Maximum value of fft bins */
	temp1, temp2;

  /* Sliding window average of bin max values */
  static int ave_max = 1000;

  /* Pointer to current pixel */
  static guchar *pix;


  /* Draw a vertical red line in waterfall at detector's freq. */
  pix = wfall_pixbuf.pixels +
	wfall_pixbuf.rowstride + wfall_pixbuf.n_channels;
  pix += wfall_pixbuf.n_channels * rc_data.red_line;
  pix[0] = 0xff;
  pix[1] = pix[2] = 0;

  /* Copy each line of waterfall to next one */
  temp1 = wfall_pixbuf.height-2;
  temp2 = wfall_pixbuf.width-2;
  for( idv = temp1; idv > 0; idv-- )
  {
	pix = wfall_pixbuf.pixels +
	  wfall_pixbuf.rowstride * idv + wfall_pixbuf.n_channels;

	for( idh = 0; idh < temp2; idh++ )
	{
	  *pix = *(pix - wfall_pixbuf.rowstride); pix++;
	  *pix = *(pix - wfall_pixbuf.rowstride); pix++;
	  *pix = *(pix - wfall_pixbuf.rowstride);
	  pix += wfall_pixbuf.n_channels - 2;
	}
  }

  /* Go to top left +1 hor. +1 vert. of pixbuf */
  pix = wfall_pixbuf.pixels +
	wfall_pixbuf.rowstride + wfall_pixbuf.n_channels;

  /* First (DC) fft bin not used */
  gbl_bin_ave[0] = 0;

  /* Do the FFT on input array */
  Ifft();
  temp1 = 2 * (wfall_pixbuf.width-2) / FFT_SIZE;
  for( fft_idx = 0; fft_idx < FFT_SIZE2; fft_idx++ )
  {
	/* Calculate vector magnitude of signal at each freq. ("bin") */
	fft_out_r[fft_idx] /= 8;
	fft_out_i[fft_idx] /= 8;
	bin_val =
	  fft_out_r[fft_idx]*fft_out_r[fft_idx] +
	  fft_out_i[fft_idx]*fft_out_i[fft_idx];

	/* Calculate average bin values */
	gbl_bin_ave[fft_idx] = ( (gbl_bin_ave[fft_idx] *
		  (BINMAX_AVE_WIN-1) + bin_val) / BINMAX_AVE_WIN );

	/* Record max bin value */
	if( bin_max < bin_val )
	  bin_max = bin_val;

	/* Scale pixel values to 255 depending on ave max value */
	pixel_val = (255 * bin_val) / ave_max;
	if( pixel_val > 255 ) pixel_val = 255;

	/* Duplicate pixels if fft size < waterfall width */
	for( idh = 0; idh < temp1; idh++ )
	{
	  /* Reverse video is toggled by right click on waterfall */
	  if( isFlagSet(REVERSE_VIDEO) )
		pix[0] = pix[1] = pix[2] = pixel_val;
	  else
		pix[0] = pix[1] = pix[2] = 255 - pixel_val;

	  pix += wfall_pixbuf.n_channels;
	}

  } /* for( fft_idx = 0; fft_idx < fft_end; fft_idx++ ) */

  /* Calculate sliding window average of max bin value */
  ave_max = (ave_max*(BINMAX_AVE_WIN-1) + bin_max)/BINMAX_AVE_WIN;
  if( !ave_max ) ave_max = 1;

  /* At last draw waterfall */
  gtk_widget_queue_draw( gbl_scope );

} /* Display_Waterfall() */

/*------------------------------------------------------------------------*/

/* Create_Pixbuff()
 *
 * Creates a GDK pixbuf for the receive window
 */
  gboolean
Create_Pixbuf( GtkWidget *drawingarea )
{
  /* Last drawing area height */
  static int
	last_wdth = 0,
	last_hght = 0;

  /* Don't create new pixbuf if size has not changed */
  if( last_hght == drawingarea->allocation.height &&
	  last_wdth == drawingarea->allocation.height )
	return( TRUE );
  last_hght = drawingarea->allocation.height;
  last_wdth = drawingarea->allocation.width;

  /* Destroy existing pixbuff */
  if( receive_pixbuf.pixbuf != NULL )
  {
	g_object_unref( G_OBJECT(receive_pixbuf.pixbuf) );
	receive_pixbuf.pixbuf = NULL;
  }

  /* Create pixbuf */
  receive_pixbuf.pixbuf = gdk_pixbuf_new(
	  GDK_COLORSPACE_RGB, FALSE, 8,
	  drawingarea->allocation.width,
	  drawingarea->allocation.height );
  if( receive_pixbuf.pixbuf == NULL ) return( FALSE );

  receive_pixbuf.pixels = gdk_pixbuf_get_pixels( receive_pixbuf.pixbuf );
  receive_pixbuf.width  = gdk_pixbuf_get_width ( receive_pixbuf.pixbuf )-1;
  receive_pixbuf.height = gdk_pixbuf_get_height( receive_pixbuf.pixbuf );
  receive_pixbuf.rowstride  = gdk_pixbuf_get_rowstride( receive_pixbuf.pixbuf );
  receive_pixbuf.n_channels = gdk_pixbuf_get_n_channels( receive_pixbuf.pixbuf );

  Clear_Pixbuf( &receive_pixbuf, gbl_drawingarea );
  return( TRUE );
} /* Create_Pixbuff() */

/*------------------------------------------------------------------------*/

/*  Clear_Pixbuf()
 *
 *  Initializes pixbuf of Receive window
 */

  void
Clear_Pixbuf( pixbuf_data_t *pixbuf, GtkWidget *drawingarea )
{
  int i, temp;
  guchar *p1, *p2;

  gdk_pixbuf_fill( pixbuf->pixbuf, 0xffffffff );

  /* Draw a box around pixbuf */
  p1 = pixbuf->pixels;
  p2 = pixbuf->pixels +
	(pixbuf->height-1) * pixbuf->rowstride;
  for( i = 0; i < pixbuf->width; i++ )
  {
	p1[0] = p1[1] = p1[2] = 0;
	p1 += pixbuf->n_channels;
	p2[0] = p2[1] = p2[2] = 0;
	p2 += pixbuf->n_channels;
  }

  p1 = pixbuf->pixels + pixbuf->rowstride;
  p2 = pixbuf->pixels + pixbuf->rowstride +
	(pixbuf->width-1) * pixbuf->n_channels;
  temp = pixbuf->height-1;
  for( i = 0; i < temp; i++ )
  {
	p1[0] = p1[1] = p1[2] = 0;
	p1 += pixbuf->rowstride;
	p2[0] = p2[1] = p2[2] = 0;
	p2 += pixbuf->rowstride;
  }

  gtk_widget_queue_draw( gbl_drawingarea );

} /* Clear_Pixbuf() */

/*------------------------------------------------------------------------*/

/*  Print_Character()
 *
 *  Prints a character to a text view port
 */

void
Print_Character(
	int printchr,
	GtkTextBuffer *text_buffer,
	GtkWidget     *scrolledwindow,
	GtkTextIter   *iter )
{
  GtkAdjustment *adjustment;
  char char2text[2];

  /* Delete a char on backspace */
  if( printchr == BS )
  {
	GtkTextIter start;

	gtk_text_buffer_get_iter_at_offset(
		text_buffer, &start,
		gtk_text_buffer_get_char_count
		(text_buffer) -1 );
	gtk_text_buffer_delete( text_buffer, &start, iter );
	return;
  }

  /* Convert char to "text" */
  char2text[0] = (char) printchr;
  char2text[1] = '\0';

  /* Print character */
  gtk_text_buffer_insert(
	  text_buffer, iter,
	  char2text, -1 );

  /* Record to QSO record file */
  if( isFlagSet(RECORD_QSO) )
	fprintf( rc_data.qso_record_fp, "%c", printchr );

  /* Scroll Text View to bottom on LF */
  if( (printchr == LF) ||
	  (printchr == GDK_Return) )
  {
	adjustment = gtk_scrolled_window_get_vadjustment
	  ( GTK_SCROLLED_WINDOW(scrolledwindow) );
	gtk_adjustment_set_value( adjustment, adjustment->upper );
	if( isFlagSet(RECORD_QSO) )
	  fprintf( rc_data.qso_record_fp, "Tx-> " );
  }

} /* Print_Character() */

/*------------------------------------------------------------------------*/

/*  Display_Signal()
 *
 *  Updates the signal scope display */

  void
Display_Signal( int plot )
{
  /* Points to plot */
  static GdkPoint *points = NULL;
  static int
	first_call = TRUE,
	points_idx = 0;


  /* Initialize on first call */
  if( first_call )
  {
	first_call = FALSE;
	points = malloc( sizeof(GdkPoint) * (wfall_pixbuf.width-2) );
  }

  /* Save values to be plotted */
  points[points_idx].y = wfall_pixbuf.height-2 - plot;
  points[points_idx].x = points_idx + 1;

  /* Recycle buffer idx when full and plot */
  if( ++points_idx == wfall_pixbuf.width-2 )
  {
	/* Draw scope background */
	gdk_draw_rectangle( gbl_scope->window,
		gbl_scope->style->white_gc,
		TRUE, 0, 0,
		wfall_pixbuf.width-1, wfall_pixbuf.height-1 );

	/* Draw scope frame */
	gdk_draw_rectangle( gbl_scope->window,
		gbl_scope->style->black_gc,
		FALSE, 0, 0,
		wfall_pixbuf.width-1, wfall_pixbuf.height-1 );

	/* Plot signal graph */
	gdk_draw_lines( gbl_scope->window,
		gbl_scope->style->black_gc,
		points, wfall_pixbuf.width-2 );

	points_idx = 0;

  } /* if( ++points_idx == SCOPE_LEN-4 ) */

} /* Display_Signal( void ) */

/*------------------------------------------------------------------------*/
